/*
 *	Simple User Authentication System -- Client Master
 *
 *	(c) 1996 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License. See file COPYING in any of the GNU utilities.
 *
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <syslog.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/wait.h>
#include <utmp.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <sys/param.h>
#include <sys/utsname.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>

#include "suas.h"

#define ERR ((char *) ~0)

char program_name[] = "suamaster";

static int sock;
static char inbuf[256];
static char master_key[80];

struct console {
  char name[8];
  pid_t pid;
};

static struct console *cons;
static int ncons;

static struct sockaddr_in remote;

static char *title_start, *title_end;

static void
title_init(char **argv, char **envp)
{
  title_start = argv[0];
  while (*envp)
	envp++;
  title_end = envp[-1] + strlen(envp[-1]);
}

static void
title_set(char *str, ...)
{
  static char buf[256];
  va_list args;
  char *c, *d;

  va_start(args, str);
  vsprintf(buf, str, args);
  va_end(args);
  c = buf;
  d = title_start;
  while (*c && d < title_end)
	*d++ = *c++;
  while (d < title_end)
	*d++ = 0;
}

static void
auth_init(char *server)
{
  struct hostent *he;

  if (!(he = gethostbyname(server)))
	{
	  syslog(LOG_ERR, "Cannot resolve server name: %s", syserr());
	  exit(1);
	}
  remote.sin_family = AF_INET;
  remote.sin_port = htons(SUA_SERVER_PORT);
  memcpy(&remote.sin_addr.s_addr, he->h_addr, he->h_length);
}

static void
auth_connect(int mode)
{
  struct sockaddr_in local;
  int one = 1;

  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock < 0)
	{
	  syslog(LOG_ERR, "socket(): %s", syserr());
	  exit(1);
	}

  if (mode)
	{
	  local.sin_family = AF_INET;
	  local.sin_port = htons(SUA_CLIENT_MASTER_PORT);
	  local.sin_addr.s_addr = INADDR_ANY;
	  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)))
		{
		  syslog(LOG_ERR, "setsockopt(): %s", syserr());
		  exit(1);
		}
	  if (bind(sock, (struct sockaddr *) &local, sizeof(local)))
		{
		  syslog(LOG_ERR, "bind(): %s", syserr());
		  exit(1);
		}
	}

  if (connect(sock, (struct sockaddr *) &remote, sizeof(remote)))
	{
	  syslog(LOG_ERR, "Cannot connect to the server: %s", syserr());
	  exit(1);
	}
}

static int accept_err;

static char *
reply(void)
{
  int z, l;
  char *b, *c;

  b = inbuf;
  l = sizeof(inbuf);
  for(;;)
	{
	  z = read(sock, b, l);
	  if (z == l)
		{
		  syslog(LOG_ERR, "Server reply too long. Confused!");
		  exit(1);
		}
	  if (!z)
		{
		  syslog(LOG_ERR, "Connection to the authentication server lost. Awaiting doom.");
		  exit(1);
		}
	  if (b[z-1] == '\r' || b[z-1] == '\n')
		break;
	  b += z;
	  l -= z;
	}
  if (b[z-1] == '\n')
	b[--z] = 0;
  if (z && b[z-1] == '\r')
	b[--z] = 0;
//  fprintf(stderr, "Reply: /%s/\n", inbuf);

  if (inbuf[0] == '-')
	{
	  if (accept_err)
		return ERR;
	  else
		{
		  syslog(LOG_ERR, "Server said: %s", inbuf);
		  exit(1);
		}
	}

  b = strchr(inbuf, '<');
  if (!b)
	return NULL;
  c = strchr(inbuf, '>');
  if (!c)
	return NULL;
  *c = 0;
  return b+1;
}

static char *
reply_get(char *to)
{
  char *k = reply();
  if (!k || k == ERR)
	{
	  syslog(LOG_ERR, "Protocol violation");
	  exit(1);
	}
  strcpy(to, k);
  return to;
}

static char *
cmd(char *cmd, ...)
{
  va_list args;
  char buf[256];
  int flag = 0;

  va_start(args, cmd);
  if (cmd[0] == '-')
	{
	  cmd++;
	  flag = 1;
	}
  vsprintf(buf, cmd, args);
  va_end(args);
//  fprintf(stderr, "Sending %s", buf);
  write(sock, buf, strlen(buf));
  return flag ? reply_get(buf) : reply();
}

static void
cons_init(char **c)
{
  int i;

  for(i=0; i<ncons; i++)
	{
	  strcpy(cons[i].name, c[i]);
	  cons[i].pid = 0;
	}
}

static void
update_wtmp(struct utmp *ut)
{
  int fd;

  if ((fd = open(_PATH_WTMP, O_APPEND | O_WRONLY)) >= 0)
	{
	  flock(fd, LOCK_EX | LOCK_NB);
	  write(fd, ut, sizeof(ut));
	  flock(fd, LOCK_UN | LOCK_NB);
	  close(fd);
	}
}

static struct utmp *
find_utmp(pid_t pid)
{
  struct utmp *ut;
  static struct utmp utt;

  utmpname(_PATH_UTMP);
  setutent();
  while ((ut = getutent()) && ut->ut_pid != pid)
	;
  if (!ut)
	{
	  ut = &utt;
	  bzero(ut, sizeof(utt));
	  ut->ut_pid = pid;
	}
  time(&ut->ut_time);
  endutent();
  return ut;
}

static void
clean_utmp(pid_t pid, char *tty, int id)
{
  struct utmp *ut = find_utmp(pid);

  if (tty)
	{
	  strcpy(ut->ut_user, "LOGIN");
	  strcpy(ut->ut_line, tty);
	  ut->ut_type = LOGIN_PROCESS;
	  sprintf(ut->ut_id, "c%d", id);
	}
  else
	{
	  bzero(&ut->ut_user, sizeof(ut->ut_user));
	  bzero(&ut->ut_host, sizeof(ut->ut_host));
	  ut->ut_type = DEAD_PROCESS;
	  ut->ut_pid = 0;
	  ut->ut_addr = 0;
	}
  pututline(ut);
  update_wtmp(ut);
  endutent();
}

static void
set_utmp(pid_t pid, char *user, char *tty)
{
  struct utmp *ut = find_utmp(pid);

  strcpy(ut->ut_line, tty);
  strcpy(ut->ut_user, user);
  ut->ut_host[0] = 0;
  ut->ut_type = USER_PROCESS;
  ut->ut_addr = 0;
  pututline(ut);
  update_wtmp(ut);
  endutent();
}

static struct termio termio;

static void
open_tty(char *name)
{
  struct stat st;
  int pid = getpid();

  close(0);
  close(1);
  close(2);
  if (chdir("/dev"))
	{
	  syslog(LOG_ERR, "chdir(\"/dev\"): %s", syserr());
	  exit(1);
	}
  if (stat(name, &st))
	{
	  syslog(LOG_ERR, "stat(%s): %s", name, syserr());
	  exit(1);
	}
  if ((st.st_mode & S_IFMT) != S_IFCHR)
	{
	  syslog(LOG_ERR, "%s: not a character device", name);
	  exit(1);
	}
  if (open(name, O_RDWR) != 0)
	{
	  syslog(LOG_ERR, "open(%s): %s", name, syserr());
	  exit(1);
	}
  if (dup(0) != 1 || dup(0) != 2)
	{
	  syslog(LOG_ERR, "dup: %s", syserr());
	  exit(1);
	}
  if (ioctl(0, TCGETA, &termio) < 0)
	{
	  syslog(LOG_ERR, "ioctl: %s", syserr());
	  exit(1);
	}
  chown(name, 0, 0);
  chmod(name, 0622);
  if (ioctl(0, TIOCSPGRP, &pid) < 0)
	syslog(LOG_ERR, "TIOCSPGRP: %s", syserr());
  ioctl(0, TCFLSH);
  termio.c_cflag |= CLOCAL | CS8 | HUPCL | CREAD | B38400;
  termio.c_iflag = termio.c_lflag = termio.c_line = 0;
  termio.c_oflag = ONLCR | OPOST;
  termio.c_cc[VMIN] = 1;
  termio.c_cc[VTIME] = 0;
  ioctl(0, TCSETA, &termio);
  fcntl(0, F_SETFL, fcntl(0, F_GETFL, 0) & ~O_NDELAY);
}

static void
issue(char *name)
{
  FILE *f;
  struct utsname uts;
  int c;

  uname(&uts);
  write(1, "\n", 1);
  if (!(f = fopen("/etc/issue", "r")))
	return;
  while ((c = getc(f)) != EOF)
	{
	  if (c == '\\')
		{
		  c = getc(f);
		  switch (c)
			{
			case 's':
			  printf("%s", uts.sysname);
			  break;
			case 'n':
			  printf("%s", uts.nodename);
			  break;
			case 'r':
			  printf("%s", uts.release);
			  break;
			case 'v':
			  printf("%s", uts.version);
			  break;
			case 'm':
			  printf("%s", uts.machine);
			  break;
			case 'o':
			  printf("%s", uts.domainname);
			  break;
			case 'l':
			  printf("%s", name);
			  break;
			default:
			  putchar(c);
			}
		}
	  else
		putchar(c);
	}
  fflush(stdout);
  fclose(f);
}

static char *
getname(int flag)
{
  static char buf[32];
  int p = 0;
  char c;

  for(;;)
	{
	  if (read(0, &c, 1) < 1)
		return NULL;
	  switch (c)
		{
		case '\r':
		case '\n':
		  goto readx;
		case 8:
		case 0x7f:
		  if (!p)
			break;
		  if (flag)
			write(1, "\b \b", 3);
		  p--;
		  break;
		case 4:
		  return NULL;
		default:
		  if (!isascii(c) || !isprint(c))
			break;
		  if (p >= sizeof(buf) - 1)
			break;
		  buf[p++] = c;
		  if (flag)
			write(1, &c, 1);
		}
	}

 readx:
  buf[p] = 0;
  return buf;
}

static void
after_tty(void)
{
#define CTRL(x) ((x) & 0x1f)

  termio.c_iflag |= IXON | IXOFF | ICRNL;
  termio.c_lflag |= ICANON | ISIG | ECHO | ECHOE | ECHOKE | ECHOK;
  termio.c_cc[VINTR] = CTRL('C');
  termio.c_cc[VQUIT] = CTRL('\\');
  termio.c_cc[VEOF] = CTRL('D');
  termio.c_cc[VERASE] = 0x7f;
  termio.c_cc[VKILL] = CTRL('U');
  if (ioctl(0, TCSETA, &termio) < 0)
	{
	  syslog(LOG_ERR, "ioctl: %s", syserr());
	  exit(1);
	}

#undef CTRL
}

static int
nologin(void)
{
  FILE *f;
  int c;

  if (!(f = fopen("/etc/nologin", "r")))
	return 0;
  while ((c = getc(f)) != EOF)
	putchar(c);
  fclose(f);
  sleep(1);
  return 1;
}

static void
threrr(char *c)
{
  fputs(c, stderr);
  sleep(1);
  exit(1);
}

static void
thread(struct console *c)
{
  char hname[MAXHOSTNAMELEN+1];
  char *name, *rep;
  char uname[32], salt[SUA_USER_SALT_SIZE+1], sess_key[SUA_SESSION_KEY_SIZE+1];
  byte uk[SUA_USER_KEY_SIZE+1], nk[SUA_NET_KEY_SIZE+1], rk[SUA_USER_KEY_SIZE+1], xbuf[256];
  struct passwd *pw;

  close(sock);
  sock = 0;
  clean_utmp(getpid(), c->name, c - cons);
  setsid();
  open_tty(c->name);
  issue(c->name);
  gethostname(hname, MAXHOSTNAMELEN);
  signal(SIGTTOU, SIG_IGN);
  for(;;)
	{
	  title_set("sua: %s", c->name);
	  write(1, hname, strlen(hname));
	  write(1, " login: ", 8);
	  name = getname(1);
	  if (!name || !name[0])
		exit(1);
	  strcpy(uname, name);
	  write(1, "\nPassword: ", 11);
	  name = getname(0);
	  write(1, "\n", 1);
	  title_set("sua: %s auth", c->name);
	  if (!sock)
		{
		  auth_connect(0);
		  reply_get(sess_key);
		}
	  strcpy(salt, cmd("-salt %s\n", uname));
	  encrypt_first(name, rk);
	  encrypt_local(rk, uk, salt);
	  encrypt_net(uk, nk, sess_key);
	  accept_err = 1;
	  rep = cmd("login %s %s %d %s\n", uname, nk, c - cons, master_key);
	  accept_err = 0;
	  if (rep != ERR)
		break;
	  write(1, "Login incorrect.\n\n", 18);
	}

  write(1, "\n", 1);
  close(sock);
  after_tty();
  if (nologin())
	exit(1);
  title_set("sua: %s logging in", c->name);
  closelog();

  pw = getpwnam(uname);
  if (!pw)
	threrr("You don't exist, go away!\n");

  setenv("HZ", "100", 1);
  setenv("TZ", "/usr/lib/zoneinfo/localtime", 1);
  set_utmp(getpid(), uname, c->name);
  if (chown(c->name, pw->pw_uid, pw->pw_gid) < 0 || chmod(c->name, 0622) < 0)
	threrr("Error chaning owner/permissions of the tty!\n");
  if (chdir(pw->pw_dir) < 0)
	threrr("Cannot find home directory!\n");
  if (setgid(pw->pw_gid) < 0)
	threrr("Cannot set primary group!\n");
  if (initgroups(pw->pw_name, pw->pw_gid) < 0)
	threrr("Cannot set secondary groups!\n");
  if (setreuid(pw->pw_uid, pw->pw_uid))
	threrr("Bad user id!\n");
  setenv("HOME", pw->pw_dir, 1);
  setenv("SHELL", pw->pw_shell, 1);
  setenv("PATH", "/bin:/usr/bin", 1);
  setenv("USER", pw->pw_name, 1);
  setenv("LOGNAME", pw->pw_name, 1);
  strcat(strcpy(xbuf, "/var/spool/mail/"), pw->pw_name);
  setenv("MAIL", xbuf, 1);

  strcpy(xbuf, "-");
  strcpy(xbuf+1, pw->pw_shell);
  execl(pw->pw_shell, xbuf, NULL);
  threrr("Cannot execute shell!\n");
}

static void
mix(pid_t pid)
{
  int i;
  pid_t p;

  for(i=0; i<ncons; i++)
	if (cons[i].pid == pid)
	  {
		if (pid)
		  clean_utmp(pid, NULL, i);
		cmd("clear %d\n", i);
		while ((p = fork()) < 0)
		  {
			syslog(LOG_ERR, "fork(): %s", syserr());
			sleep(1);
		  }
		if (p)
		  cons[i].pid = p;
		else
		  {
			thread(cons+i);
			exit(0);
		  }
	  }
}

static volatile int perint;

static void
alarm_handler(int eger)
{
  perint = 1;
}

static struct sigaction sigact = { alarm_handler, 0, 0, NULL };

void
main(int argc, char **argv, char **envp)
{
  openlog(argv[0], LOG_NDELAY | LOG_PID, LOG_DAEMON);

  close(0);
  close(1);
  close(2);
  setsid();
  sigaction(SIGALRM, &sigact, NULL);

  if (argc < 3)
	{
	  syslog(LOG_ERR, "Invalid arguments");
	  exit(1);
	}
  ncons = argc - 2;
  cons = xmalloc(sizeof(struct console) * ncons);
  auth_init(argv[1]);
  auth_connect(1);
  reply_get(master_key);
  cmd("master\n");
  cons_init(argv + 2);

  title_init(argv, envp);
  title_set("suamaster");

  mix(0);
  alarm(SUA_REFRESH);

  for(;;)
	{
	  int stat;
	  pid_t pid;

	  pid = wait(&stat);
	  if (perint)
		{
		  perint = 0;
		  cmd("nop\n");
		  alarm(SUA_REFRESH);
		}
	  if (pid > 0)
		{
		  alarm(0);
		  mix(pid);
		}
	  alarm(SUA_REFRESH);
	}
}
